%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                   The C-IFF Proof Procedure                     %%%
%%%        for Abductive Logic Programming with Constraints         %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Auxiliary predicates (file ciff-aux.pl)                         %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% v1.3: 30/11/2003 by Ulle Endriss (ue@doc.ic.ac.uk)              %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% This file provides a number of auxiliary predicates that are not
%%% specific to the implementation of the C-IFF proof procedure.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% +Term =** [-Functor|-Arguments].
%%% Given a term, return its functor and its arguments as a list.
%%% Works like the built-in operator "=..", but fails whenever Term
%%% is a variable (rather than causing an instantiation error).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- op( 700, xfx, =**).

Term =** [Functor|Arguments] :-
  \+ var( Term),
  Term =.. [Functor|Arguments].


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% subterm( +Term1, +Term2).
%%% Succeed if Term1 is  a term appearing as a proper (and exact) 
%%% subexpression within Term2.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

subterm( T1, T2) :-
  T2 =** [_|Ts],
  member( T, Ts),
  (T == T1 ; subterm( T1, T)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% exact_member( +Elem, +List).
%%% Succeed if an exact copy of the term Elem can be found in List.
%%% This predicate is useful, for instance, to check whether a given
%%% variable belongs to a certain list of variable names. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

exact_member( Elem, [Head|_]) :-
  Elem == Head.

exact_member( Elem, [_|Tail]) :-
  exact_member( Elem, Tail).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% exact_subset( +List1, +List2).
%%% Succeed if an exact copy of every element in List1 can be found 
%%% in List2.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

exact_subset( [], _).

exact_subset( [Head|Tail], List) :-
  exact_member( Head, List),
  exact_subset( Tail, List).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% delete_list( +List, +DeleteList, -Result).
%%% Given two lists List and DeleteList, Result will be instantiated
%%% with the result of deleting From List any elements in DeleteList.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

delete_list( List, [], List).

delete_list( List, [D|DelList], Result) :-
  delete( List, D, List1),
  delete_list( List1, DelList, Result).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% subst( +[X1,...,Xn], +[T1,...,Tn], +Formula, -Formula1).
%%% Given a list of variables [X1,...,Xn] and a list of terms
%%% [T1,...,Tn] of the same length (such that no Ti contains any of
%%% the Xi) together with an expression Formula, return the 
%%% expression we get by replacing every occurrence of Xi in Formula
%%% with Ti (for all i), i.e. Formula1 = Formula[X1/T1]...[Xn/Tn].
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

subst( [X|Xs], [T|Ts], Formula, Formula1) :-
  subst( X/T, Formula, Formula2),
  subst( Xs, Ts, Formula2, Formula1).

subst( [], _, Formula, Formula).

subst( X/T, Y, T) :- 
  X == Y, !.

subst( X/_, Y, Y) :-
  var( Y),
  \+ (X == Y), !.

subst( _/_, A, A) :-
  ground( A), !.

subst( X/T, A, A1) :-
  A =.. [P|Ts],
  subst_list( X/T, Ts, Ts1),
  A1 =.. [P|Ts1].

subst_list( X/T, [Head|Tail], [Head1|Tail1]) :-
  subst( X/T, Head, Head1),
  subst_list( X/T, Tail, Tail1).

subst_list( _, [], []).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% listify( +List, -LList).
%%% Given a list such as [a,b,c], return a list of singleton lists
%%% such as [[a],[b],[c]].
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

listify( [], []).

listify( [Head|Tail], [[Head]|Tail1]) :-
  listify( Tail, Tail1).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% construct_equalities( +Terms1, +Terms2, -Eqs).
%%% Given two lists of terms Terms1 = [t1..tk] and Terms2 = [s1..sk] 
%%% (of the same length), return the list of eqalities [t1=s1..tk=sk].
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

construct_equalities( [], _, []).

construct_equalities( [Head1|Tail1], [Head2|Tail2], [Head|Tail]) :-
  Head = (Head1 = Head2),
  construct_equalities( Tail1, Tail2, Tail).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% filter( +List, +Functor, -Sublist).
%%% Get all elements from List that start with Functor.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

filter( [Head|Tail], Functor, [Head|Tail1]) :-
  functor( Head, Functor, _), !,
  filter( Tail, Functor, Tail1).

filter( [_|Tail], Functor, List) :-
  filter( Tail, Functor, List).

filter( [], _, []).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% read_listing( +File, -List).
%%% Read a the Prolog terms from a given File into a List.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

read_listing( File, List) :-
  open( File, read, Stream),
  read_stream( Stream, List).  

read_stream( Stream, [Head|Tail]) :-
  read( Stream, Head),
  \+ (Head == end_of_file), !,
  read_stream( Stream, Tail).

read_stream( _, []).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
