%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                   The C-IFF Proof Procedure                     %%%
%%%        for Abductive Logic Programming with Constraints         %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Constraint handling (file ciff-constraints.pl)                  %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% v1.3: 30/11/2003 by Ulle Endriss (ue@doc.ic.ac.uk)              %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% This file provides a wrapper around the  Sicstus Prolog built-in 
%%% constraint solver for finite domain CLP constraints. In principle,
%%% any other solver could be used, as long as it implements the 
%%% predicates given here.
%%%
%%% In the current implementation, the following constraints are 
%%% admissible: X#=Y, X#\=Y, X#<Y, X#=<Y, X#>Y, X#>=Y, where X and Y
%%% are arithmetic expressions involving variables and integeres.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Load Sicstus Prolog library for CLP finite domain constraints.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module( library(clpfd)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% set_domain_min( +N).
%%% set_domain_max( +N).
%%% Set the minimal (maximal) value (a "small" integer) for any 
%%% constraint variable in the constraint store. The defaults are 0 
%%% and 100, respectively.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_domain_min( N) :-
  retractall( domain_min(_)),
  assert( domain_min(N)).

:- set_domain_min( 0).

set_domain_max( N) :-
  retractall( domain_max(_)),
  assert( domain_max(N)).

:- set_domain_max( 100).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% post_constraint( +Constraint).
%%% Post a Constraint to the constraint store, i.e. call it. 
%%% In addition, for every variable occurring in Constraint, a domain
%%% constraint is also posted to the store.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

post_constraint( Constraint) :-
  term_variables( Constraint, Vars),
  domain_min( Min),
  domain_max( Max),
  domain( Vars, Min, Max),
  call( Constraint).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% post_constraints( +Constraints).
%%% Post every constraint in the list Constraints.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

post_constraints( []).

post_constraints( [C|Cs]) :-
  post_constraint( C),
  post_constraints( Cs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% exhaustive_check( Vars).
%%% Perform an exhaustive check for satifiability of the current 
%%% constraint with respect to a given list of variables. Complete but
%%% computationally expensive.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

exhaustive_check( Vars) :-
  once( labeling( [], Vars)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% is_constraint( +Term).
%%% Succeed if Term represents a constraint.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

is_constraint( _ #= _).
is_constraint( _ #\= _).
is_constraint( _ #< _).
is_constraint( _ #>= _).
is_constraint( _ #> _).
is_constraint( _ #=< _).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% complementary_constraint( +Constraint, -Complement).
%%% Given a Constraint, return its Complement.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

complementary_constraint( X #= Y, X #\= Y).
complementary_constraint( X #\= Y, X #= Y).
complementary_constraint( X #< Y, X #>= Y).
complementary_constraint( X #>= Y, X #< Y).
complementary_constraint( X #> Y, X #=< Y).
complementary_constraint( X #=< Y, X #> Y).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% equality_constraint( +X, +Y, -Constraint).
%%% Construct the equality constraint with argument X and Y.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

equality_constraint( X, Y, X #= Y).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% disequality_constraint( +X, +Y, -Constraint).
%%% Construct the disequality constraint with argument X and Y.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disequality_constraint( X, Y, X #\= Y).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% simplify_constraints( +Constraints, -NewConstraints).
%%% Simplify a given list of constraints. This is not complete and 
%%% mostly for the purpose of presentation. The following steps are
%%% being performed:
%%% (1) Reorientation: rewrite X#>Y as Y#<X etc.
%%% (2) Remove duplicate constraints.
%%% (3) Remove constraints that are subsumed by a *single* other
%%%     constraint (e.g. X#=<Y is subsumed by X#<Y, and N1#<X is 
%%%     subsumed by N2#<X if N2 is a greater number than N1).
%%% (4) Remove redundant constraints such as X#=X.
%%% Note that trivial constraints (without variables) are already 
%%% being removed by the main procedure.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

simplify_constraints( Constraints, NewConstraints) :-
  reorient_constraints( Constraints, Constraints1),
  remove_duplicates( Constraints1, Constraints2),
  remove_subsumed_constraints( Constraints2, Constraints2, Constraints3),
  remove_redundant_constraints( Constraints3, NewConstraints).


reorient_constraints( [], []).

reorient_constraints( [X#>Y|Constraints], [Y#<X|NewConstraints]) :- !,
  reorient_constraints( Constraints, NewConstraints).

reorient_constraints( [X#>=Y|Constraints], [Y#=<X|NewConstraints]) :- !,
  reorient_constraints( Constraints, NewConstraints).

reorient_constraints( [C|Constraints], [C|NewConstraints]) :- 
  reorient_constraints( Constraints, NewConstraints).


remove_subsumed_constraints( [], _, []).

remove_subsumed_constraints( [C|Constraints], FullList, NewConstraints) :-
  member( C1, FullList),
  subsumed( C, C1), 
  (subsumed( C1, C) -> exact_member( C1, Constraints) ; true), !,
  remove_subsumed_constraints( Constraints, FullList, NewConstraints).

remove_subsumed_constraints( [C|Constraints], FullList, [C|NewConstraints]) :-
  remove_subsumed_constraints( Constraints, FullList, NewConstraints).


subsumed( X1#=<Y1, X2#=Y2)  :- X1 == X2, Y1 == Y2.
subsumed( X1#=<Y1, X2#<Y2)  :- X1 == X2, Y1 == Y2.
subsumed( X1#\=Y1, X2#<Y2)  :- X1 == X2, Y1 == Y2.

subsumed( X1#=Y1, Y2#=X2)   :- X1 == X2, Y1 == Y2.
subsumed( X1#\=Y1, Y2#\=X2) :- X1 == X2, Y1 == Y2.

subsumed( X1#<N1, X2#<N2)   :- X1 == X2, number( N1), number( N2), N2 < N1.
subsumed( X1#=<N1, X2#=<N2) :- X1 == X2, number( N1), number( N2), N2 < N1.
subsumed( X1#=<N1, X2#<N2)  :- X1 == X2, number( N1), number( N2), N2 =< N1+1.
subsumed( X1#<N1, X2#=<N2)  :- X1 == X2, number( N1), number( N2), N2 < N1-1.

subsumed( N1#<X1, N2#<X2)   :- X1 == X2, number( N1), number( N2), N1 < N2.
subsumed( N1#=<X1, N2#=<X2) :- X1 == X2, number( N1), number( N2), N1 < N2.
subsumed( N1#=<X1, N2#<X2)  :- X1 == X2, number( N1), number( N2), N1 =< N2+1.
subsumed( N1#<X1, N2#=<X2)  :- X1 == X2, number( N1), number( N2), N1 < N2-1.


remove_redundant_constraints( [], []).

remove_redundant_constraints( [C|Constraints], NewConstraints) :- 
  member( C, [X#=Y,X#=<Y]),
  X == Y, !,
  remove_redundant_constraints( Constraints, NewConstraints).

remove_redundant_constraints( [C|Constraints], [C|NewConstraints]) :-  
  remove_redundant_constraints( Constraints, NewConstraints).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
