%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                   The C-IFF Proof Procedure                     %%%
%%%        for Abductive Logic Programming with Constraints         %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Main program (file ciff-main.pl)                                %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% v1.3: 30/11/2003 by Ulle Endriss (ue@doc.ic.ac.uk)              %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% To run the C-IFF proof procedure, use either ciff/3 or ciff/4. 
%%% Use switch_debugging/1 to change the output of debugging 
%%% information. 
%%%
%%% The proof procedure itself is implemented in ciff-proc.pl, 
%%% ciff-constraints.pl provides the constraint solver, and a number 
%%% of auxiliary predicates are implemented in ciff-aux.pl.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% This implementation is (in part) based on the following paper:
%%% T.H. Fung and R. Kowalski. The IFF Proof Procedure for Abductive
%%% Logic Programming. Journal of Logic Programming, 33(2):151-165,
%%% 1997. Available at http://www.doc.ic.ac.uk/~rak/.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Syntax of abductive logic programs with constraints:
%%%
%%% Definitions: A iff Ds
%%%  - where A is a "normal" atom all of whose arguments are variables 
%%%    and Ds is a list of lists of literals, representing a 
%%%    disjunction of conjunctions of literals
%%%
%%% Integrity constraints (ICs): Cs implies Ds
%%%  - where Cs is a list of literals and Ds is a list of atoms,
%%%    representing a conjunction and a disjunction, respectively
%%%
%%% Queries: Cs
%%%  - a list of literals, representing a conjunction
%%%
%%% Literals are either atoms or negative literals.
%%%
%%% Negative literals: not(A) or [A] implies [false]
%%%  - where A is an atom
%%%
%%% Atoms may be "normal" atoms such as p(X,bob), equalities such as 
%%% X = Y, or constraints.
%%%
%%% Constraints: see ciff-constraints.pl
%%%
%%% An ALP is a list of definitions and integrity constraints.
%%% 
%%% Queries: a list of literals, representing a conjunction
%%%
%%% Allowedness conditions: 
%%%  - A definition is allowed iff any variable occurring either in a
%%%    constraint in a disjunct on the righthand side or anywhere in a
%%%    disjunct on the righthand side but not on the lefthand side also
%%%    occurs in a non-constraint/non-equality atom in that disjunct.
%%%  - An integrity constraint is allowed iff any variable occurring 
%%%    anywhere in the integrity constraint also occurs in a 
%%%    non-constraint atom in the antecedent.
%%%  - A query is allowed iff any variable occurring anywhere in the
%%%    query also occurs in a non-constraint atom in the query.
%%%
%%% Files: may contain any sequence of definitions and ICs
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Load Sicstus Prolog library for lists.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module( library(lists)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Load Sicstus Prolog library for term utilities.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module( library(terms)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Declare required propositional connectives as operators.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- op( 400, yfx, implies),
   op( 500, yfx, iff).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Consult the other files of the program.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- consult( 'ciff-aux.pl').
:- consult( 'ciff-proc.pl').
:- consult( 'ciff-constraints.pl').


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% switch_debugging( +Status).
%%% Status should be either "on", "step", "off", or "dots". If 
%%% debugging is switched "on" then all proof steps are printed out. 
%%% If the status is "step" then Prolog will stop after each proof step
%%% and wait for the user to hit the return key. If the status is set 
%%% to "dots" then the system will simply print out a single dot for 
%%% each proof step. The default is "off". To change the type of output 
%%% etc., see also predicate inform/3.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

switch_debugging( Status) :-
  retractall( debugging(_)),
  assert( debugging(Status)).

:- switch_debugging( off).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% ciff( +File, +Query, -Answer).
%%% ciff( +Defs, +ICs, +Query, -Answer).
%%% Run the C-IFF procedure given the ALP in File (or represented by
%%% the supplied Definitions and ICs) for the given Query and return
%%% an Answer. Any proposition without an explicit iff-definion in Defs
%%% is assumed to be abducible. Answer is of the form A:S:C where A is 
%%% the set of abducible atoms, S is the substitution, and C is a list 
%%% of constraints. The set of all answers can be collected by forced
%%% backtracking.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ciff( File, Query, Answer) :-
  read_alp( File, Defs, ICs),
  ciff( Defs, ICs, Query, Answer).

ciff( Defs, ICs, Query, Answer) :-
  preprocess( Defs, Defs1),
  preprocess( ICs, ICs1),
  preprocess( Query, Query1),
  append( Query1, ICs1, Node),
  term_variables( Query, FreeVars),
  inform( 'initialise', '', Node),
  sat( Node, FreeVars, []:[], [], Defs1, FreeVars, Answer).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% read_alp( +File, -Defs, -ICs).
%%% Get the ALP consisting of iff-Definitions and ICs from File.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

read_alp( File, Defs, ICs) :-          
  read_listing( File, Listing),
  filter( Listing, iff, Defs),
  filter( Listing, implies, ICs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% preprocess( +ALP, -NewALP).
%%% Rewrite a given ALP (a list of iff-definitions and integrity using
%%% the internal syntax. In the current implementation this only 
%%% involves rewriting any occurrence of not(X) as [X] implies [false].
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

preprocess( X, X) :-
  var( X), !.

preprocess( A, A) :-
  atomic( A), !.

preprocess( not(Term), [NewTerm] implies [false]) :- !,
  preprocess( Term, NewTerm).

preprocess( [Head|Tail], [NewHead|NewTail]) :- !,
  preprocess( Head, NewHead),
  preprocess( Tail, NewTail).

preprocess( Term, NewTerm) :-
  Term =.. [Functor|Args],
  preprocess( Args, NewArgs),
  NewTerm =.. [Functor|NewArgs].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
